# Copyright 2023-2024 The Wazo Authors  (see the AUTHORS file)
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from copy import copy
from types import MethodType
from typing import Optional

import requests
from cachetools.func import ttl_cache
from jwt import DecodeError, ExpiredSignatureError
from jwt import decode as jwt_decode
from wazo_auth.exceptions import UnknownExternalAuthConfigException
from wazo_auth.flask_helpers import Tenant
from wazo_auth.http import required_acl
from wazo_auth.plugins.external_auth.mobile.http import MobileAuthSenderID
from wazo_auth_client import Client as AuthClient
from xivo.rest_api_helpers import APIException

logger = logging.getLogger(__name__)


class JWTRefreshError(Exception):
    pass


class ConfigurationError(Exception):
    pass


class PushServicesUnavailable(APIException):
    def __init__(self):
        msg = 'Push services unavailable, try again later'
        super().__init__(503, msg, 'push-services-unavailable', {}, 'external')


class Plugin:
    CACHE_TTL: float = 3600  # 1h

    def __init__(self):
        self._jwt: Optional[str] = None
        self._last_known_sender_id: Optional[str] = None

    def load(self, dependencies: dict) -> None:
        try:
            self._configure(dependencies)
        except ConfigurationError as e:
            logger.error(e)
            raise

        logger.info(
            'Patching %s:%s to support Wazo Mobile Push Proxy',
            MobileAuthSenderID.__module__,
            MobileAuthSenderID.__name__,
        )
        MobileAuthSenderID.get = MethodType(self._patched_get, MobileAuthSenderID)

    def _configure(self, dependencies: dict) -> None:
        depends_keys = {'config', 'external_auth_service'}
        if missing_keys := (depends_keys - dependencies.keys()):
            raise ConfigurationError(
                f'Missing dependencies: {", ".join(missing_keys) }'
            )

        self._external_auth_service = dependencies['external_auth_service']
        config: dict = dependencies['config']

        config_keys = {'nestbox', 'mobile_push_proxy_host', 'mobile_push_proxy_port'}
        if missing_keys := (config_keys - config.keys()):
            raise ConfigurationError(
                f'Missing keys in configuration: {", ".join(missing_keys)}'
            )

        nestbox_config: dict
        if not (nestbox_config := copy(config['nestbox'].get('auth', {}))):
            raise ConfigurationError('Nestbox: no authentication section found')

        if 'username' not in nestbox_config:
            try:
                nestbox_config['username'] = nestbox_config.pop('service_id')
            except KeyError:
                raise ConfigurationError('Nestbox: no username found')

        if 'password' not in nestbox_config:
            try:
                nestbox_config['password'] = nestbox_config.pop('service_key')
            except KeyError:
                raise ConfigurationError('Nestbox: no password found')

        self._proxy_host = config['mobile_push_proxy_host']
        self._proxy_port = config['mobile_push_proxy_port']
        self._nestbox_params = nestbox_config

    @required_acl('auth.users.{user_uuid}.external.mobile.sender_id.read')
    def _patched_get(self, instance_cls: MobileAuthSenderID, user_uuid: str):
        tenant: Tenant = Tenant.autodetect()

        sender_id = self._sender_id_from_mobile_config(tenant.uuid)

        if not sender_id:
            sender_id, fetch_error = self._sender_id_from_push_proxy()

        if not sender_id:
            sender_id = self._sender_id_from_last_known()

        if not sender_id:
            raise fetch_error

        return {'sender_id': sender_id}, 200

    def _refresh_jwt(self) -> str:
        try:
            jwt_decode(
                self._jwt,
                algorithms=['RS256'],
                options={
                    'verify_signature': False,
                    'verify_exp': True,
                    'verify_aud': False,
                },
            )
        except (AttributeError, ExpiredSignatureError, DecodeError):
            client = AuthClient(**self._nestbox_params)
            try:
                token = client.token.new(expiration=10)
                self._jwt = token['metadata']['jwt']
            except requests.RequestException as e:
                raise JWTRefreshError('Failed to refresh JWT from nestbox') from e
            except KeyError:
                raise JWTRefreshError('No JWT received with token')
        return self._jwt

    def _sender_id_from_mobile_config(self, tenant_uuid: str) -> Optional[str]:
        try:
            return self._external_auth_service.get_config('mobile', tenant_uuid).get(
                'fcm_sender_id', None
            )
        except UnknownExternalAuthConfigException:
            return None

    def _sender_id_from_push_proxy(self):
        try:
            return self._fetch_sender_id(), None
        except (JWTRefreshError, PushServicesUnavailable) as e:
            return None, e

    def _sender_id_from_last_known(self):
        if self._last_known_sender_id:
            logger.info('using last known value of sender-id')
            return self._last_known_sender_id
        return None

    @ttl_cache(ttl=CACHE_TTL, maxsize=1)
    def _fetch_sender_id(self) -> str:
        jwt = self._refresh_jwt()

        headers = {'Authorization': f'key={jwt}'}
        url = f'https://{self._proxy_host}:{self._proxy_port}/fcm/sender-id'
        logger.debug('Requesting `sender-id` from %s', url)

        try:
            response = requests.get(url, headers=headers, timeout=10)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error('Failed to connect: %s', e)
            raise PushServicesUnavailable()

        sender_id = self._last_known_sender_id = response.json().get('sender_id')
        return sender_id
